import json
import os
from json import JSONDecodeError
from typing import Dict, Optional

import requests
from requests import Response


def download_json(url: str, headers: Dict[str, str] = None):
    headers = headers or {}

    try:
        r: Response = requests.get(url, headers=headers, timeout=20)
    except requests.exceptions.ConnectionError:
        raise

    if r.status_code != 200:
        raise Exception(f"HTTP error: code[{r.status_code}], message: {r.reason}")
    else:
        try:
            j = json.dumps(r.json())
        except JSONDecodeError:
            msg = f"Cannot decode JSON from '{url}'"
            raise Exception(f"JSONDecodeError: {msg}")
        return j


def interpret_pypi(json_string):
    rj = json.loads(json_string)
    pypi_version = rj["info"]["version"]
    return pypi_version


def interpret_devpi(json_string):
    rj = json.loads(json_string)
    versions = list(rj["result"])

    def try_int(x):
        try:
            return int(x)
        except ValueError:
            return 0

    version_sorted = sorted(versions, key=lambda _: tuple(map(try_int, _.split("."))))

    v = version_sorted[-1]
    return v


def get_pip_server():
    default = "https://pypi.org/pypi"
    r = os.environ.get("PIP_INDEX_URL", default)
    if r.endswith("/"):
        r = r[:-1]
    return r


def get_last_version_fresh(package_name: str) -> Optional[str]:
    pip_server = get_pip_server()
    # logger.info(pip_server=pip_server)
    if "pypi.org" in pip_server:
        pip_server = "https://pypi.org/pypi"
        url = f"{pip_server}/{package_name}/json"
        headers = {"Accept": "application/json"}
        # noinspection PyBroadException
        try:
            j = download_json(url, headers=headers)
        except Exception:
            return None
        return interpret_pypi(j)
    else:
        headers = {"Accept": "application/json"}
        url = f"{pip_server}/{package_name}"
        # noinspection PyBroadException
        try:
            j = download_json(url, headers=headers)
        except Exception:
            return None

        return interpret_devpi(j)



import os
from typing import Optional

import requirements
from requirements.requirement import Requirement

from .versions import get_last_version_fresh

__all__ = ["update_versions"]


def update_versions(fin: str, fout: Optional[str] = None):
    fout = fout or fin + ".resolved"

    if not os.path.exists(fin):
        print(f"Requirement file '{fin}' does not exist")
        return

    print(f"Now updating requirements files '{fin}' -> '{fout}'")

    resolved = []
    with open(fin) as fd:
        req: Requirement
        for req in requirements.parse(fd):
            todo = (len(req.specs) == 0) or \
                   (len(req.specs) == 1 and req.specs[0][0] in (">=", ">"))
            if todo:
                v = get_last_version_fresh(req.name)
                if v is None:
                    msg = f"Cannot find the package '{req.name}'"
                    raise Exception(msg)

                print(f"Updated package: {req.name} to {v}")
                req.specs = [("==", v)]
            else:
                print(f"Skipping spec {req}: {req.specs}")

            ss = ",".join(f"{a}{b}" for a, b in req.specs)
            s2 = f"{req.name}{ss}"
            resolved.append(s2)

    # compile output file
    res = "\n".join(resolved)
    # write file to disk
    with open(fout, "wt") as buf:
        buf.write(res)



#!/usr/bin/env python3

import argparse

from ..update import update_versions


def resolve():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-i",
        "--input",
        type=str,
        required=True,
        help="Input dependencies file"
    )

    parser.add_argument(
        "-o",
        "--output",
        type=str,
        required=False,
        help="Output (resolved) dependencies file"
    )

    parsed = parser.parse_args()

    update_versions(fin=parsed.input, fout=parsed.output)


if __name__ == '__main__':
    resolve()
